#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"
#
#
#                M O V I E D A T A . T C L
#
# Command: moviedata.tcl
#
# Synopyis:
#         moviedata.tcl  <data_file>[.movie_data.csv]
#
# Desciption:
#     Display the details of how the errors in object positions will change over the
#     duration of the exposure.  This uses the movie data output file produced
#     by the "moviedata" command - a CVS file with the details we need.
#
#     This script reads the file and displays two GUI windows.  One displays details
#     from the data file and allows a summary of each time point to be displayed.
#
#     The second GUI window displays the field of view and the position of the
#     reference position, optimized position and the object at the current time
#     point.  You can (will need to) zoom in on a given object and show how it
#     moves whilst time-points are changed.
#
# Author: Tony Farrell, AAO
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata.tcl,v 12.10 03-Oct-2014 08:40:38+10 tjf $"
#
# History:
#     22-Oct-2008 - TJF - Intial experimental version.
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata.tcl,v 12.10 03-Oct-2014 08:40:38+10 tjf $"
#
# Implementation Notes:
#    The moviedata_reader.tcl script contains the code needs to read the CVS file
#    into a Tcl array variable, the name of which is passed in ("Data"). 
#
#    The moviedata_gui.tcl script contains the code to build the main GUI window.
#
#    The moviedata_mimic.tcl script contains the code to build the mimic.
#

if { $argc != 1 } {
    puts stderr "$argv0:Usage $argv0 <moveidata csv file>"
    exit 1
}
set file [lindex $argv 0]
#puts "File is $file"

if { [ file readable $file ] == 0 } {
    set f2 "${file}.movie_data.csv"
    if { [ file readable $f2 ] == 0 } {
        puts stderr "$argv0:Can't read move data file\n    Tried \"$file\" and \"$file.movie_data.csv\""
        exit 1
    }
    set file $f2
}


# Source the code which can read our data file.
source moviedata_reader.tcl

# Source the code which will construct our GUI.
source moviedata_gui.tcl
source moviedata_mimic.tcl


# and read it.
ReadDataFile $file Data

# Contruct and run the GUI.
ConstructGui "" Data

