#                M O V I E D A T A _ M I M I C . T C L
#
# Script: moviedata_mimic.tcl
#
# Description:
#     Used by moviedata_gui.tcl to implement the GUI Mimic window.
#
#     Entry point is "ContructMimic"
#
#     Data file should have been read by moviedata_reader.tcl into the
#     variable the name of which is passed in.
#
# Author: Tony Farrell, AAO
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_mimic.tcl,v 12.10 03-Oct-2014 08:40:35+10 tjf $"
#
# History:
#     22-Oct-2008 - TJF - Intial experimental version.
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_mimic.tcl,v 12.10 03-Oct-2014 08:40:35+10 tjf $"
##
# Notes:
#   Currently use sizes in microns (size of object etc).  Data file should have
#     conversion from microns to arc-seconds and we should use arc-seconds.
#
#   Implementation is a bit messy - needs some tidying up.  Somewhat based
#     on configure mimic code.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


#
# This method will contruct the mimic used to display the movie data
# as the time points are selected.
#
proc ConstructMimic { data } {

    upvar #0 $data d

    toplevel .mimic

    # Work out the window size.
    set mimicSize [WindowSize .mimic]

    # Radius in microns we must represent - must be floating point.
    set radiusMicrons [expr $d(MaxFieldRadius)*1.0]




    frame .mimic.info   -padx 2 -borderwidth 2 -relief ridge
    frame .mimic.mimic  -padx 2 -borderwidth 2 -relief ridge

    pack .mimic.info .mimic.mimic -expand 1 -fill both

    MimicConstructMimic .mimic.mimic $data $mimicSize $radiusMicrons
    MimicConstructInfo  .mimic.info $data .mimic.mimic.field
}

#
# Build the info area of the mimc.
#
proc MimicConstructInfo { w data mw } {
    
    upvar #0 $data d
    #
    set row 0


    label $w.zoomLab -text "Zoom Factor:"
    entry $w.zoomEnt -textvariable ${data}(NewZoom)
    bind $w.zoomEnt <Key-Return> "DoZoom $data $mw"

    grid $w.zoomLab -row $row -column 0
    grid $w.zoomEnt -row $row -column 1

    incr row
    label $w.refSizeLab -text "Ref Pos Size" -fg $d(RefPos,color)
    entry $w.entRefPos -textvariable ${data}(RefPos,NewSize)

    label $w.optimSizeLab -text "Opt Pos Size" -fg $d(OptimPos,color)
    entry $w.entOptimPos -textvariable ${data}(OptimPos,NewSize)

    label $w.objectSizeLab -text "Object Size" -fg $d(Object,color)
    entry $w.entObject -textvariable ${data}(Object,NewSize)

    label $w.sizeLab -text "(All Microns)"

    grid $w.refSizeLab $w.entRefPos $w.optimSizeLab $w.entOptimPos  \
        $w.objectSizeLab $w.entObject $w.sizeLab 

    incr row

    
    # Set up bindinds for each of the above.
    foreach t { RefPos OptimPos Object } {
        
        bind $w.ent${t} <Key-Return>  "MimicObjSizeSet $t $data"
    }



    label $w.fibNumLab -text "Fibre Num" 
    label $w.fibNumVal -border 2 -relief sunken -textvariable ${data}(CurFib,Num)

    label $w.fibTypeLab -text "Type"
    label $w.fibTypeVal -border 2 -relief sunken  -textvariable ${data}(CurFib,Type)

    label $w.fibPosLab -text "Position"
    label $w.fibPosVal -border 2 -relief sunken  -textvariable ${data}(CurFib,Pos)

    grid $w.fibNumLab $w.fibNumVal $w.fibTypeLab  $w.fibTypeVal $w.fibPosLab  $w.fibPosVal -row $row -sticky ew

    incr row

    label $w.errLab -text "Distance of object from"
    label $w.errRefLab -text "Ref Pos"
    label $w.errRefVal -border 2 -relief sunken  -textvariable ${data}(ErrFromRef)

    label $w.errOptLab -text "Optimal Pos"
    label $w.errOptVal -border 2 -relief sunken  -textvariable ${data}(ErrFromOptim)


    grid $w.errLab  -column 1 -row $row -sticky ew
    grid $w.errRefLab -column 2 -row $row -sticky ew
    grid $w.errRefVal -column 3 -row $row -sticky ew
    grid $w.errOptLab -column 4 -row $row -sticky ew
    grid $w.errOptVal -column 5 -row $row -sticky ew
    incr row
}

#
# Invoked to set the object size. tag indicates the type of the objects.
#
proc MimicObjSizeSet { tag data  } {
    upvar #0 $data d


    #puts "MimicObjSizeSet invoked - tag $tag"

    # Fetch the suggested new value.
    set newVal $d($tag,NewSize)

    # Validate the value.
    if { [string is integer -strict $newVal] != 1 } {
        set d($tag,NewSize) $d($tag,Size)
        error "The  value of $newVal is not a positive integer value"
    } elseif { ($newVal < 1) || ($newVal > 1000) } {
        set d($tag,NewSize) $d($tag,Size)
        error "The value of $newVal is outside the range 1 to 1000"
    }

    # Set the new value and update mimic.
    set d($tag,Size) $d($tag,NewSize)

    MimicUpdate $data

}

#
# Contruct the actual mimic area.
#
proc MimicConstructMimic { w data wSize radiusMicrons } {


    upvar #0 $data d

    canvas $w.field -height $wSize -width $wSize  \
        -yscrollcommand "$w.vscroll set" \
        -xscrollcommand "$w.hscroll set" \
        -scrollregion "0 0 $wSize $wSize" \
        -xscrollincrement 20 -yscrollincrement 20 -background white
       

    scrollbar $w.hscroll -orient horizontal  -command "set ${data}(Clicked) 0 ; $w.field xview"


    scrollbar $w.vscroll -orient vertical -command "set ${data}(Clicked) 0 ;$w.field yview"

    pack $w.vscroll -side right -fill y 
    pack $w.field -side top -expand 1
    pack $w.hscroll   -fill x

    # Initialise a bunch of variables used throughout.
    set d(Zoom) 1.0
    set d(NewZoom) 1.0
    set d(MimicSize) $wSize
    set d(RadiusMicrons) $radiusMicrons

    # 
    # Values here should be in arc-seconds, not mircons.  But this is
    # not yet available in the moviedata file.
    #
    #  Values below are for 2dF.  
    #     Scale - 70 microns/arc-sec.
    #     Fibre size - 1.5 arc-seconds - 105 microns
    #     Object size - 1 arc-second - 70 microns.
    #
    set d(RefPos,Size) 105
    set d(OptimPos,Size) 105
    set d(Object,Size) 70

    set d(RefPos,NewSize) 105
    set d(OptimPos,NewSize) 105
    set d(Object,NewSize) 70

    set d(RefPos,color) yellow
    set d(OptimPos,color) green
    set d(Object,color) purple

    set d(FieldWindow) $w.field

    set d(CurFib,Num) ""
    set d(CurFib,Type) ""
    set d(CurFib,Pos) ""

    set d(ErrFromRef) ""
    set d(ErrFromOptm) ""

    set d(Clicked) 0

#   Set up mouse wheel bindings if we have not yet done so
    bind all <Button-5> {Dtclu___WheelEvent %X %Y 5 units}
    bind all <Button-4> {Dtclu___WheelEvent %X %Y -5 units}
    bind all <Control-Button-5> {Dtclu___WheelEvent %X %Y 1 pages}
    bind all <Control-Button-4> {Dtclu___WheelEvent %X %Y -1 pages}


    # Set bindings on clicking.
    bind $w.field <ButtonPress-1> "SetClickPosition $data %x %y"

    bind $w.field <ButtonPress-3> " \
        SetClickPosition $data %x %y ;\
        ZoomBy $data $w.field 4.0 "

    bind $w.field <Control-ButtonPress-3> "\
        SetClickPosition $data %x %y ;\
        ZoomBy $data $w.field 0.5 "
        
    

    # Trigger initial data display.
    DoZoom $data $w.field 1.0 
}


#
# Draw points of a given tag (Original/Optimized position)
#
proc MimicDrawPoints { data w tag } {
    upvar #0 $data d

    # Radius of circle to draw.
    set radius [expr $d($tag,Size)/2]

    set color $d($tag,color)

    for { set i 1 } { $i <= $d(numTotal) } { incr i } {
        set x $d(Fibre,$i,X,$tag)
        set y $d(Fibre,$i,Y,$tag)

        set bottomLeftX [CvtXMicronsToPixels $data [expr -1 * $radius + $x]]
        set bottomLeftY [CvtYMicronsToPixels $data [expr -1 * $radius + $y]]
        set topRightX   [CvtXMicronsToPixels $data [expr      $radius + $x]]
        set topRightY   [CvtYMicronsToPixels $data [expr      $radius + $y]]

        #puts "$w create oval $bottomLeftX $bottomLeftY $topRightX $topRightY -tag $tag -fill $color"
        $w create oval $bottomLeftX $bottomLeftY $topRightX $topRightY \
            -tag $tag -fill $color \
            -tag object_${tag}_$i

        $w bind object_${tag}_$i <Enter> "OverObject $data $tag $i 0"
        $w bind object_${tag}_$i <Leave> "LeftObject $data $tag $i 0"
        
    }


    
}


#
# Draw object position at the current time point.
#
proc MimicDrawObjectsAtTimePoint { data w } {

    upvar #0 $data d

    # Radius of circle to draw.
    set radius [expr $d(Object,Size)/2]

    set color $d(Object,color)



    set point $d(CurrentTimePointNum)

    for { set i 1 } { $i <= $d(numTotal) } { incr i } {
        set x $d(Fibre,$i,X,$point)
        set y $d(Fibre,$i,Y,$point)

        set bottomLeftX [CvtXMicronsToPixels $data [expr -1 * $radius + $x]]
        set bottomLeftY [CvtYMicronsToPixels $data [expr -1 * $radius + $y]]
        set topRightX   [CvtXMicronsToPixels $data [expr      $radius + $x]]
        set topRightY   [CvtYMicronsToPixels $data [expr      $radius + $y]]

        $w create oval $bottomLeftX $bottomLeftY $topRightX $topRightY \
            -tag TimePointPos -fill $color \
            -tag object_tp_$i

        $w bind object_tp_$i <Enter> "OverObject $data TimePointPos $i $point"
        $w bind object_tp_$i <Leave> "LeftObject $data TimePointPos $i $point"
        
    }




}
#
# Update the mimic.
#
proc MimicUpdate { data } {
    upvar #0 $data d

    set w $d(FieldWindow)
    $w delete all



    set radius $d(RadiusMicrons)


    set bottomLeftX [CvtXMicronsToPixels $data [expr -1 * $radius]]
    set bottomLeftY [CvtYMicronsToPixels $data [expr -1 * $radius]]
    set topRightX   [CvtXMicronsToPixels $data [expr      $radius]]
    set topRightY   [CvtYMicronsToPixels $data [expr      $radius]]


#    puts "radius = $radius"
#    puts "$w create oval $bottomLeftX $bottomLeftY $topRightX $topRightY "


#   Draw the edge of the displayed area.
    $w create oval $bottomLeftX $bottomLeftY $topRightX $topRightY \
            -tag Boundry -width 2

#   Draw the reference and optimized positions.
    MimicDrawPoints  $data $w RefPos 
    MimicDrawPoints  $data $w OptimPos 

#   Draw the object at the current time point.
    MimicDrawObjectsAtTimePoint $data $w
  
}


#
# Invoked to increase the zoom by the specified factor.
#
proc ZoomBy { data w by } {

    #puts "ZoomBy invoked $data $w $by"

    upvar #0 $data d
    set z [expr $d(Zoom) * $by]
    if { $z > 1000 } {
        set z 1000
        bell
    } elseif { $z < 0.5 } {
        set z 0.5
        bell
    }
    DoZoom $data $w $z
}

# Actually zoom the mimic.
#
# data = where we store the application data
# w = mimic window.
# zc = zoom to apply.
#
#
proc DoZoom { data w { zc ""} } {

    upvar #0 $data d


    # If specified, Save the zoom value. Otherwise zoom to the value in
    #  the NewZoom variable.
    if { $zc != "" } {
        set d(NewZoom) $zc
    } else {
        set zc $d(NewZoom)
    }
    if { [string is double -strict $zc] != 1 } {
        set d(NewZoom) $d(Zoom)
        error "The zoom value of $zc is not a positive floating point value"
    } elseif { ($zc < 0.1) || ($zc > 1000) } {
        set d(NewZoom) $d(Zoom)
        error "The zoom value of $zc is outside the range 0.1 to 1000"
    }

    set d(Zoom) $zc


    #  Get the currently displayed region limits. 
    #  SetWindowSize{} will reset these.
    
    set XViewDescr [$w xview]
    set YViewDescr [$w yview]

    set XStartFraction [lindex $XViewDescr 0]
    set YStartFraction [lindex $YViewDescr 0]
    set XEndFraction [lindex $XViewDescr 1]
    set YEndFraction [lindex $YViewDescr 1]

    if { $d(Clicked) } {
        set XCenter $d(XClickPosition)
        set YCenter $d(YClickPosition)
    } else {
        #  Zoom center point  - we zoom about the center of the current
        #  display.

        set XCenter [ expr ($XStartFraction + $XEndFraction) * 0.5 ]
        set YCenter [ expr ($YStartFraction + $YEndFraction) * 0.5 ]
    }

    # Work out the center in X and Y
    set xc [expr $XCenter - (1.0 / ($zc * 2.0))]
    set yc [expr $YCenter - (1.0 / ($zc * 2.0))]

    #  work ut the scroll region limit and set it.
    set z [expr $zc * $d(MimicSize)]
    $w configure -scrollregion "0 0 $z $z"

    # Set the scroll position
    $w xview moveto $xc
    $w yview moveto $yc

    # Update the display.
    MimicUpdate $data

}

#
# Invoked when we move the mouse over an object.  Display the values for that object.
#
proc OverObject { data tag fibNum timePoint } {
    if { $tag == "RefPos" } {
        OverRefPos $data $fibNum
    } elseif { $tag == "OptimPos" } {
        OverOptimPos $data $fibNum
    } elseif { $tag == "TimePointPos" } {
        OverTimePointPos $data $fibNum $timePoint
    } else {
        puts "OverObject invoked with unknown tag $tag"
    }
}

#
# Invoked when mouse moves off an object.  Clear the displayed values.
#
proc LeftObject { data tag fibNum timePoint } {
    upvar #0 $data d
    set d(CurFib,Index) ""
    set d(CurFib,Num) ""
    set d(CurFib,Type) ""
    set d(CurFib,Pos) ""
    set d(ErrFromRef) ""
    set d(ErrFromOptim) ""
}


#
# We are over a reference position - display the values for it.
#
proc OverRefPos { data i } {
    upvar #0 $data d
    set d(CurFib,Index) $i
    set d(CurFib,Num) $d(Fibre,$i,Num)
    set d(CurFib,Type) "Ref"
    set d(CurFib,Pos) "$d(Fibre,$i,X,RefPos), $d(Fibre,$i,Y,RefPos)"

    UpdateErrors $data $i
}
#
# We are over an optimized position - display the values for it.
#
proc OverOptimPos { data i } {
    upvar #0 $data d

    set d(CurFib,Index) $i
    set d(CurFib,Num) $d(Fibre,$i,Num)
    set d(CurFib,Type) "Optimized"
    set d(CurFib,Pos) "$d(Fibre,$i,X,OptimPos), $d(Fibre,$i,Y,OptimPos)"

    UpdateErrors $data $i

}
#
# We are over the object at the specified time point - display values for it.
#
proc OverTimePointPos { data i timePoint } {
    upvar #0 $data d
    #puts "Fibre $d(Fibre,$i,Num):PosAtTimePoint: X:$d(Fibre,$i,X,$timePoint), Y:$d(Fibre,$i,Y,$timePoint)"
    set d(CurFib,Index) $i
    set d(CurFib,Num) $d(Fibre,$i,Num)
    set d(CurFib,Type) "Actual at TP"
    set d(CurFib,Pos) "$d(Fibre,$i,X,$timePoint), $d(Fibre,$i,Y,$timePoint)"

    UpdateErrors $data $i
}

#
# Display the error between the object position at this time point and the reference
# and optimized positions
#
proc UpdateErrors { data i } {
    upvar #0 $data d
    set timePoint  $d(CurrentTimePointNum)

    set xDiff [expr $d(Fibre,$i,X,$timePoint) -$d(Fibre,$i,X,RefPos)]
    set yDiff [expr $d(Fibre,$i,Y,$timePoint) -$d(Fibre,$i,Y,RefPos)]

    set d(ErrFromRef)   [format "%.1f" [expr sqrt($xDiff*$xDiff + $yDiff*$yDiff)]]

    set xDiff [expr $d(Fibre,$i,X,$timePoint) -$d(Fibre,$i,X,OptimPos)]
    set yDiff [expr $d(Fibre,$i,Y,$timePoint) -$d(Fibre,$i,Y,OptimPos)]

    set d(ErrFromOptim)   [format "%.1f" [expr sqrt($xDiff*$xDiff + $yDiff*$yDiff)]]

}
#                   S e t  C l i c k  P o s i t i o n
#
#   Records the given position as the 'last clicked on' position. This
#   is the last position in the mimic display that was clicked on, and
#   so is a sensible place to zoom around. The X and Y positions passed
#   should be in canvas pixel coordinates. This means they will be from
#   0.0 to the window size in pixels in X and Y.
proc SetClickPosition { data X Y } {
    #puts "SetClickPosition $data $X $Y"

    upvar #0 $data d

    
    #  The canvas xview command returns two fractions. The canvas displays
    #  a portion of the scroll region associated with it. The first fraction
    #  is the fraction of the scroll region to the left of the displayed
    #  area, and the second is the fraction of the scroll region to the
    #  left of the right edge of the display. That is, the two fractions give
    #  the display limits in X expressed as fractions of the whole scroll
    #  region.
    #
    #  The display is XWindowSize pixels by YWindowSize pixels (at the moment,
    #  both of these are the same and are obtained from $d(mimicSize)).
    #  We make sure these are in floating point, so calculations are all
    #  done in floating point.
    #
    #  We determine the start and end fractions of the scroll region currently
    #  displayed, and calculate the fraction of the scroll region actually
    #  shown in the display.
    
    set XWindowSize [expr $d(MimicSize) * 1.0]
    set YWindowSize [expr $d(MimicSize) * 1.0]

    set w $d(FieldWindow)

    set XViewDescr [$w xview]
    set XStartFraction [lindex $XViewDescr 0]
    set XEndFraction [lindex $XViewDescr 1]

    set YViewDescr [$w yview]
    set YStartFraction [lindex $YViewDescr 0]
    set YEndFraction [lindex $YViewDescr 1]
    

    set XDisplayedFraction [expr $XEndFraction - $XStartFraction]
    set YDisplayedFraction [expr $YEndFraction - $YStartFraction]
    
    #  From that we can calculate the X and Y Click positions as fractions
    #  of the scroll region.
    
    set XClickPosition [expr (($X / $XWindowSize) * $XDisplayedFraction) \
                                                        + $XStartFraction]
    set YClickPosition [expr (($Y / $YWindowSize) * $YDisplayedFraction) \
                                                        + $YStartFraction]
    
    #  And, just to be sure, we check they're in the range 0..1
    
    if { $XClickPosition < 0.0 } { set XClickPosition 0.0 }
    if { $XClickPosition > 1.0 } { set XClickPosition 1.0 }
    if { $YClickPosition < 0.0 } { set YClickPosition 0.0 }
    if { $YClickPosition > 1.0 } { set YClickPosition 1.0 }
    
    #  Set the global variables that record the scrollbar setting when
    #  this routine was called.
  
    set d(XClickPosition)    $XClickPosition
    set d(YClickPosition)    $YClickPosition
    set d(XClickScrollStart) $XStartFraction
    set d(YClickScrollStart) $YStartFraction

    # Used to indicate we have clicked, since last scroll.
    set d(Clicked) 1

}


#
# Convert value in microns to pixels.
#
proc CvtXMicronsToPixels { data xMicrons } {
    upvar #0 $data d
    set centre $d(MimicSize)/2
    return [expr $d(Zoom) *  \
            (($d(MimicSize) / $d(RadiusMicrons) * $xMicrons / 2.0) + $centre)]
}
proc CvtYMicronsToPixels { data yMicrons } {
    upvar #0 $data d
    set centre $d(MimicSize)/2
    return [expr $d(Zoom) * \
            (($d(MimicSize) / $d(RadiusMicrons) * $yMicrons / 2.0) + $centre)]
}

#
# Work out the window size ot use.
#
proc WindowSize { w } {

    # We use 80% of the maximum size we have available.
    set x [winfo screenwidth  $w]
    set y [winfo screenheight $w]

    if { $x < $y } {
        set size [expr int($x * 0.8)]
    } else {
        set size [expr int($y * 0.8)]
    }

    

}

#
# Get the radius in microns for the mimic.
#
proc RadiusMicrons { data } {

    upvar #0 $data d

    # 
    # 
    # Work through the fibres and determine the maximum radius we need.  The
    # reference position of each fibre is sufficient.
    #
    #  Actually would be better if we new the real size of the insrument (E.g 2dF
    # field plate diameter)
    #
    set radius 0
    for { set i 1 } { $i <= $d(numTotal) } { incr i } {
        set x $d(Fibre,$i,X,RefPos)
        set y $d(Fibre,$i,Y,RefPos)

        set r [expr sqrt($x*$x + $y*$y)]

        if { $r > $radius } {
            set radius $r
        }
    }
    
    return [expr $radius + 10000]

}



#
# This procedure is to be invoked for Button4 and Button5 events.  It
# checks if the window in which we got the event has a vertical scroll bar.
# If yes, then it scrolls that bar by "delta" "units".  x and y are the
# location of the event.
#
# This procedure is a significant modificaton of one available from
#
#   http://aspn.activestate.com/ASPN/Cookbook/Tcl/Recipe/68394 which
#
# is noted on that site as being in the public domain. (and not under
#  GNU licence).
#
#
proc Dtclu___WheelEvent { x y  delta unit} {

    # Find out what's the widget we're on
    set act 0
    set widget [winfo containing $x $y]

    if {$widget != ""} {
        # Make sure we've got a vertical scrollbar option for this widget
        if {[catch "$widget cget -yscrollcommand" cmd]} return

        # Scroll the widget (We do this regardless of if we have
        #  the scrollbar)
        $widget yview scroll $delta $unit

    }
}

